#include "config.h"
#include "stddefs.h"
#include "cpuio.h"
#include "genlib.h"
#include "cache.h"
#include "warmstart.h"
#include "cpu.h"
#include "cdefBF537.h"

#define THR		0x00
#define RBR		0x00
#define DLL		0x00
#define DLH		0x04
#define IER		0x04
#define LCR		0x0c
#define MCR		0x10
#define LSR		0x14
#define SCR		0x1c
#define GCTL	0x24

#define UART0_BASE	0xffc00400
#define UART1_BASE	0xffc02000
#define CONSOLE(reg) (UART0_BASE+reg)

/* If for some reason, DMA-based character reception on the UARTs is
 * not desireable, then comment out this definition...
 */
#define DMA_UART_RCV

#if PLATFORM_VTCEZ
#undef DMA_UART_RCV
#endif

#ifdef DMA_UART_RCV
#define UBUFSIZ 32
uint32_t ubuf0[UBUFSIZ/4];
uint8_t *uptr0;
uint8_t *uend0;
#endif


struct abcd {
	int a;
	char b;
	int c;
	long d;
};

struct abcd S1, S2;

void
test_func(int mode)
{
	S1.a = 44;
	S1.b = 'c';
	S1.c = 9999;
	S1.d = 0x10002345;
	printf("S1: %d 0x%02x %d 0x%lx\n",S1.a, S1.b, S1.c, S1.d);

	if (mode == 1) {
		printf("This will fail...\n");
		S2 = S1;
	}
	else {
		printf("This will work...\n");
		memcpy((char *)&S2,(char *)&S1,sizeof(struct abcd));
	}

	printf("S2: %d 0x%02x %d 0x%lx\n",S2.a, S2.b, S2.c, S2.d);
}

/* devInit():
 * Initialize console...
 */
void
devInit(int baud)
{
	ushort tmp;

	*(vushort *)CONSOLE(GCTL) = 1;		/* Enable UART clocks */
	*(vushort *)CONSOLE(MCR) = 0;
	*(vushort *)CONSOLE(LCR) = 0x03;	/* no parity, 8-bit char, 1 stopbit */
	tmp = *(vushort *)CONSOLE(LSR);		/* clear LSR */
	*(vushort *)CONSOLE(IER) = 0;		/* no ints */
	*(vushort *)CONSOLE(SCR) = 0;		/* no ints */
	ConsoleBaudSet(baud);

#ifdef DMA_UART_RCV
	/* Init DMA8 (UART0 RCV):
	 */
	*pDMA8_START_ADDR = ubuf0;
	*pDMA8_CURR_ADDR = ubuf0;
	*pDMA8_X_COUNT = UBUFSIZ;
	*pDMA8_X_MODIFY = 1;
	*pDMA8_CURR_X_COUNT = 0;
	*pDMA8_CONFIG = DMAEN | WNR | WDSIZE_8 | FLOW_AUTO;
	*pUART0_IER = ERBFI;
	uptr0 = (uint8_t *)ubuf0;
	uend0 = uptr0+UBUFSIZ;
#endif

}

/* ConsoleBaudSet():
 * Change baud rate to that of 'baud'.
 */
int
ConsoleBaudSet(int baud)
{
	unsigned short	lctl, hi, lo;

	lo = SYSCLKFREQ/(baud*16);
	hi = (lo & 0xff00) >> 8;
	lo &= 0x00ff;

	lctl = *(vushort *)CONSOLE(LCR);	/* Save linectl reg */
	*(vushort *)CONSOLE(LCR) = DLAB;
	*(vushort *)CONSOLE(DLL) = lo;		/* Set baud divisor */
	*(vushort *)CONSOLE(DLH) = hi;
	*(vushort *)CONSOLE(LCR) = lctl;	/* Restore linectl reg */
	return(0);
}

/* target_console_empty():
 * Returns 1 if there are no characters waiting to
 * be put out on the UART; else return 0 indicating that the UART
 * is still busy outputting characters from its FIFO.
 */
int
target_console_empty(void)
{
	if (*(ushort *)CONSOLE(LSR) & TEMT)
		return(1);
	return(0);
}

/* target_putchar():
 * When buffer has space available, load the incoming character
 * into the UART.
 */
int
target_putchar(char c)
{
    /* Wait for transmit ready bit */
	while(!(*(vushort *)CONSOLE(LSR) & THRE));

	*(vushort *)CONSOLE(THR) = (ushort)c;

    return((int)c);
}

/* target_gotachar():
 * Return 0 if no char is avaialable at UART rcv fifo; else 1.
 * Do NOT pull character out of fifo, just return status. 
 */
int
target_gotachar(void)
{
#ifdef DMA_UART_RCV
	if (*pDMA8_CURR_ADDR != uptr0)
		return(1);
#else
	if (*(vushort *)CONSOLE(LSR) & DR)
		return(1);
#endif
	return(0);
}

/* target_getchar():
 * Assume there is a character in the UART's input buffer and just
 * pull it out and return it.
 */
int 
target_getchar(void)
{
#ifdef DMA_UART_RCV
	if (uptr0 == uend0)
		uptr0 = (uint8_t *)ubuf0;
	return(*uptr0++);
#else
    return((int)*(vushort *)CONSOLE(RBR));
#endif
}

int
iInvalidateRange(char *start, int size)
{
	long addr, end, imemctrl;


	/* If start and size are zero, then we are to invalidate the entire
	 * i-Cache.  That's done by clearing the IMC bit then re-setting it.
	 */
	if ((start == 0) && (size == 0)) {
		imemctrl = *pIMEM_CONTROL;

		asm("ssync;");
		*pIMEM_CONTROL &= ~IMC;
		asm("csync;");
		*pIMEM_CONTROL = imemctrl;
		asm("csync;");
	}
	else {
		addr = (long)start;
		addr &= ~0x1f;
		end = (long)(start + size);

		for(;addr < end; addr += 32)
			IFlush((char *)addr);
	}
	return(0);
}

int
dFlushRange(char *start, int size)
{
	long addr, end;

	/* TODO:
	 * This needs to support the "start=0,size=0" input to flush
	 * the entire d-cache.
 	 */
	addr = (long)start;
	addr &= ~0x1f;
	end = (long)(start + size);

	for(;addr < end; addr += 32)
		Flush((char *)addr);

	return(0);
}

/* cacheInitForTarget():
 * Establish target specific function pointers and
 * enable i-cache...
 * Refer to $core/cache.c for a description of the function pointers.
 * NOTE:
 * If cache (either I or D or both) is enabled, then it is important
 * that the appropriate cacheflush/invalidate function be established.
 * This is very important because programs (i.e. cpu instructions) are
 * transferred to memory using data memory accesses and could
 * potentially result in cache coherency problems.
 */
void
cacheInitForTarget(void)
{
	extern int icache_off(void);
	extern int dcache_off(void);

	icacheDisable = icache_off;
	dcacheDisable = dcache_off;
	dcacheFlush = dFlushRange;
	icacheInvalidate = iInvalidateRange;
}

/* target_reset():
 * From BF537 programmer's manual...
 * The reset vector is determined by the processor system. It points to the
 * start of the on-chip boot ROM (0xef000000), or to the start of
 * external asynchronous memory (0x20000000), depending on the state of
 * the BMODE pins.  uMon is built to run in Mode 0, which means the CPU
 * starts fetching from location 0x20000000 immediately out of reset.
 */
void
target_reset(void)
{
	flushDcache(0,0);
	disableDcache();
	invalidateIcache(0,0);
	disableIcache();

	/* Use Watchdog to generate reset...
	 */
	*pWDOG_CTL = 0x0ad0;	/* Disable watchdog */
	*pWDOG_CNT = 0x0004;	/* Set a low count value */
	*pWDOG_CTL = 0x0000;	/* Enable watchdog */
	while(1);				/* Wait for the hit */
}

void
ledloop(void)
{
	volatile int i;

	*pPORTFIO_DIR |= 0x0100;
	while(1) {
		*pPORTFIO_SET |= 0x0100;
		for(i=0;i<1000000;i++);
		*pPORTFIO_CLEAR |= 0x0100;
		for(i=0;i<1000000;i++);
	}
}

/* If any CPU IO wasn't initialized in reset.S, do it here...
 * This just provides a "C-level" IO init opportunity. 
 */
void
initCPUio(void)
{
	/* ADD_CODE_HERE */
}

int
bfin_i2cshow(int interface)
{
#if PLATFORM_VTCEZ
	printf("Addr      Description\n");
	printf("0x42      ADV7173B Multiformat SDTV Video Decoder\n");
	printf("0xb8      MT9V022  Digital Image Sensor\n");
#else
	printf("SHOW function not available on BFIN eval board.\n");
#endif
	return(0);
}
