//=============================================================================
//
//      cpuio.c
//
//      CPU/Board Specific IO
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         05-26-2002
// Description:  This file contains the IO functions required by Micro Monitor
//				 that are unique to each CPU/Board combination
//
//
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"
#include "uart16550.h"
#include "pxa250.h"
#include "warmstart.h"

extern ulong i2c_init();
extern void ps2_kbd_init(void);
extern ulong ps2_gotachar(void);
extern uchar ps2_getchar(void); 
extern int GPIO_clr(int);
extern void GPIO_init(void);
extern void lcd_putchar(char c);
extern void lcd_init(void);
extern void lcd_on(void);

extern int lcd_color_depth;
extern void putpsr(ulong);
extern ulong getpsr(void);

// Delay for some usecs. - Not accurate, assumes ROM mode
// and no Cache
void udelay(int delay)
{
  int i;
  for ( i = LOOPS_PER_USEC * delay; i ; i--)
	;
}

void
cf_init(void)
{
	extern int GPIO_tst(int);
	extern int GPIO_clr(int);
	extern int GPIO_set(int);

	// read the CD signals to see if we have a card
	// both must be 0
	if (GPIO_tst(8) || GPIO_tst(9)){
#ifdef CF_DEBUG
		printf("No Compact FLASH Card Detected.\n");
#endif
		return;
	}

#ifdef CF_DEBUG
	printf("Compact FLASH Card Detected...");
#endif

	// setup the PCMCIA/CF timing registers for maximum delay
	// this is necessary until the card CIS can be read and
	// actual operating parameters determined
	MSC_REG(MSC_MCMEM0) = 0x000fcfff;
	MSC_REG(MSC_MCMEM1) = 0x000fcfff;
	MSC_REG(MSC_MCATT0) = 0x000fcfff;
	MSC_REG(MSC_MCATT1) = 0x000fcfff;
	MSC_REG(MSC_MCIO0) = 0x000fcfff;
	MSC_REG(MSC_MCIO1) = 0x000fcfff;

	MSC_REG(MSC_MECR) = MECR_CIT;		// 1 socket

	// enable power to the card - 3.3V ONLY!
	GPIO_clr(11);

	// delay and then check the error flag
	udelay(10000);

	if (GPIO_tst(12) == 0){
		GPIO_set(11); // turn off the power!
		printf("Power Error, Removing Card Power!\n");
		return;
	}
	printf("Power OK!\n");
}

int
getUartDivisor(int baud, unsigned char *hi, unsigned char *lo)
{
	// The PXA250 derives it's baud clock from the 14.7456Mhz input clock
	*lo = ((14745600/16)/baud) & 0x00ff;
	*hi = (((14745600/16)/baud) & 0xff00) >> 8;
	return(0);
}

int
devInit(int baud)
{

	lcd_color_depth = 8;

	// initialize the GPIO
	GPIO_init();

	// Initialize the UART
    InitUART(baud);

	// Initialize the I2C Port
	i2c_init();

	// INitialize the LCD controller
	lcd_init();
	lcd_on();

	// Turn on LED0 and 1
	GPIO_clr(3);
	GPIO_clr(4);

#if INCLUDE_PS2KBD
	// Initialize the PS2 controller and keyboard - if attached
	ps2_kbd_init();
#endif

	// check for the presence of a CF card
	cf_init();

	return 0;
}

void
initCPUio()
{
}

void
intsrestore(ulong psr)
{
    putpsr(psr);
}

/*
 * Read the program status register (CPSR)
 * and set the FIQ and IRQ bits.
 */
ulong
intsoff()
{
    ulong  psr;

    psr = getpsr();

    /*
     * Set bit 6, bit 7 to disable interrupts.
     */
    putpsr(psr | 0x000000c0);
    return(psr);
}

/* cacheInitForTarget():
    Enable instruction cache only...
*/
void
cacheInitForTarget(void)
{
    asm("   MRC p15, 0, r0, c1, c0, 0");
    asm("   ORR r0, r0, #0x1000");  /* bit 12 is ICACHE enable*/
    asm("   MCR p15, 0, r0, c1, c0, 0");

    /* Flush instruction cache */
    asm("   MCR p15, 0, r0, c7, c5, 0");
}

void
target_reset(void)
{
	warmstart(INITIALIZE);
}
