/* Monitor configuration file.
    General notice:
    This code is part of a boot-monitor package developed as a generic base
    platform for embedded system designs.  As such, it is likely to be
    distributed to various projects beyond the control of the original
    author.  Please notify the author of any enhancements made or bugs found
    so that all may benefit from the changes.  In addition, notification back
    to the author will allow the new user to pick up changes that may have
    been made by other users after this version of the code was distributed.

    Author: Ed Sutter
    email:  esutter@lucent.com
    phone:  908-582-2351
*/

#ifndef CONFIG_H
#define CONFIG_H

/* Establish default MAC/IP addresses...
 * This MAC address is owned by COGENT Computer Systems.
 */
#define     DEFAULT_ETHERADD "00:30:23:00:00:01"
#define     DEFAULT_IPADD 	 "192.168.1.110"

#define MONSTACKSIZE (8*1024)

/* Specify CPU/PLATFORM type and name so that common code can be used
 * for a similar cpu, on different platforms.
 * The 'TYPE' definition is used for ifdefs in the code and the 'NAME'
 * is used for printfs in the code.
 */
#define CPU_NAME        "PowerPC 405GP"
#define PLATFORM_NAME   "Cogent CSB272"

#define RAM_VECTOR_TABLE    0x0

#define PRE_COMMMANDLOOP_HOOK()	flash_lock()

/* Defining DONT_CENTER_MONHEADER will eliminate the automatic centering
 * of the monitor's startup banner...
 */
#define DONT_CENTER_MONHEADER

/* Base-address of boot flash must be hard-coded to 0xfff80000
 */
#define BOOTROMBASE_OVERRIDE 0xfff80000

/* LOOPS_PER_SECOND:
 * Approximately the size of a loop that will cause a 1-second delay.
 * This can be guestimated or modified with the sleep -c command at the
 * monitor command line.
 */

/* This value calibrated for instruction cache on, data cache on,
 * ethernet polling enabled.
 */
#define LOOPS_PER_SECOND    140000

/* The CSB272 hardware simply has one bank of flash devices.
 * We define SINGLE_FLASH_DEVICE here to pull in the default
 * code from the flash driver file.
 */
#define SINGLE_FLASH_DEVICE		

/* Flash bank configuration:
 * This target is configured with one flash bank.
 */
#define FLASHRAM_BASE			0x200000
#define FLASHRAM_END			0x2fffff
#define FLASHRAM_SECTORSIZE		0x010000
#define FLASHRAM_SECTORCOUNT	16
#define FLASHRAM_BANKNUM		1

#ifdef FLASHRAM_BASE
#define FLASHBANKS              2
#else
#define FLASHBANKS              1
#endif

#define FLASH_BANK0_WIDTH       4
#define FLASH_BANK0_BASE_ADDR	0xfe000000
#define FLASH_LARGEST_SECTOR    0x40000
#define FLASH_TIMEOUT			1000000
#define FLASH_COPY_TO_RAM		1

/* TFS definitions:
 *  TFSSTART:       Base address in FLASH at which TFS starts.
 *  TFSEND:         End address of TFS in FLASH.
 *  TFSSPARE:       Location of sector that is used as the spare sector
 *                  by TFS for defragmentation.
 *  TFSSPARESIZE:   Size of the spare sector used by TFS for defragmentation.
 *  TFSSECTORCOUNT: Number of eraseable sectors that TFS covers, not including
 *                  the TFSSPARE sector.
 */
#define TFSSPARESIZE			FLASH_LARGEST_SECTOR
#define TFSSTART				0xff000000
#define TFSEND					0xfff3ffff
#define TFSSPARE				(TFSEND+1)
#define TFSSECTORCOUNT			((TFSSPARE-TFSSTART)/0x40000)
#define TFS_EBIN_ELF			1
#define TFS_VERBOSE_STARTUP		1
#define TFS_AUTOBOOT_ABORTABLE	    

#define FLASH_PROTECT_RANGE  "126-127"

/* SYMFILE is the name of a file that, if present, the monitor uses to 
 * process symbols on the command line.  A symbol is a whitespace delimited
 * string prefixed by a percent sign (%).
 */
#define SYMFILE "symtbl"

/* Specify the size of the memory block (in monitor space) that is to be
 * allocated to malloc in the monitor.  Note that this size can be dynamically
 * increased through the HEAPBASE and HEAPSTART shell variables.
 */
#define ALLOCSIZE 64*1024

#define NO_MONLIB_PCI_STUBS
#define NO_MONLIB_I2C_STUBS
#define USE_DEFAULT_PCISHOW

/* Specify inclusion of subsystems within the monitor here.
 */
#define INCLUDE_MEMCMDS         1
#define INCLUDE_EDIT            1
#define INCLUDE_DISASSEMBLER    1
#define INCLUDE_UNZIP           1
#define INCLUDE_ETHERNET        1
#define INCLUDE_ICMP			1
#define INCLUDE_TFTP            1
#define INCLUDE_DHCPBOOT        1
#define INCLUDE_TFS             1
#define INCLUDE_TFSCLI          1
#define INCLUDE_TFSAPI          1
#define INCLUDE_TFSSYMTBL       1
#define INCLUDE_TFSSCRIPT       1
#define INCLUDE_XMODEM          1
#define INCLUDE_LINEEDIT        1
#define INCLUDE_EE              1
#define INCLUDE_FLASH           1
#define INCLUDE_MALLOC          1
#define INCLUDE_SHELLVARS       1
#define INCLUDE_STRACE          1
#define INCLUDE_CAST            1
#define INCLUDE_STRUCT          1
#define INCLUDE_REDIRECT        1
#define INCLUDE_QUICKMEMCPY     1
#define INCLUDE_PROFILER        1
#define INCLUDE_BBC             1
#define INCLUDE_MEMTRACE        1
#define INCLUDE_SPRCMD          1
#define INCLUDE_DCRCMD          1
#define INCLUDE_TRAPCMD         1
#define INCLUDE_SYSCALLCMD      1
#define INCLUDE_STOREMAC		0
#define INCLUDE_PS2KBD			0
#define INCLUDE_SED				0
#define INCLUDE_VERBOSEHELP		1
#define INCLUDE_HWTMR			0
#define INCLUDE_PORTCMD			0
#define INCLUDE_SYSLOG			1
#define INCLUDE_USRLVL			1
#define INCLUDE_PCI				1

#if INCLUDE_PS2KBD
#define MORE_PUTCHAR	sed_putchar
#define MORE_GOTACHAR	ps2_gotachar
#define MORE_GETCHAR 	ps2_getchar
#endif

#define CONSOLE_UART_BASE	0xef600300	/* See UART1_BASE in cpuio.h */

/* Inclusion of this next file will make sure that all of the above
 * inclusions are legal; and adjust where necessary.
 */
#include "inc_check.h"

#endif  /* CONFIG_H */
