    .file "reset.s"

/*
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author:  Ed Sutter
 * email:   esutter@lucent.com
 * phone:   908-582-2351
 * 
 *
 * Modified for the CSB431 - ML67Q500x/400x (ARM7TDMI) Single Board
 * Mike Kelly, mike@cogcomp.com
 *
 * rom_reset.s:
  */

#include "config.h"
#include "warmstart.h"
#include "ml67q5003.h"

    /*
     * Have a separate stack for each processor mode.
     */

    /* define sizes for each mode's stack */
    .equ FiqStackSz, 4096
    .equ IrqStackSz, 4096
    .equ AbtStackSz, 4096
    .equ UndStackSz, 4096
    .equ SysStackSz, 4096

    /* declare the stacks */
    .extern MonStack
    .global FiqStack
    .global IrqStack
    .global AbtStack
    .global UndStack
    .global SysStack        
    
    /* allocate the stacks */
    .comm   FiqStack, FiqStackSz    /* for the FIQ mode */
    .comm   IrqStack, IrqStackSz    /* for the IRQ mode */
    .comm   AbtStack, AbtStackSz    /* for the Abort mode */
    .comm   UndStack, UndStackSz    /* for the Undef mode */
    .comm   SysStack, SysStackSz    /* for the System mode */
    /* User mode has the same stack as system mode. */

/*********************************************************************/
    
    .extern start

    .global reset
    .global coldstart
    .global lukewarmstart
    .global warmstart
    .global etheraddr
    .global moncomptr

    .text
        
    /*
     * Exception table at address 0
     */ 
reset:  
    b coldstart
    b undefined_instruction
    b software_interrupt
    b abort_prefetch
    b abort_data
    b not_assigned
    b interrupt_request
    b fast_interrupt_request

/********************************************************************/
    
    /* Provide space to allow a programmer to place an ASCII 
     * string in this location as an optional point
     * of storage for MAC address...
     */ 

etheraddr:
    .byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
    .byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
    .byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff

moncomptr:
    .long   moncom

/*********************************************************************/

    /*
     * At the end of the reset sequence, MMU, Icache, Dcache,
     * and write buffer are all disabled.
     * Also IRQs and FIQs are disabled in the processor's CPSR
     * The operating mode is SVC (supervisory mode), and the
     * PC is vectored at 0x00000000. A branch in 0x00000000
     * brings us directly here.
	 *
	 */

coldstart:
	ldr	pc, =coldstart_1		/* jump to actual ROM location */
	nop

coldstart_1:
	bl sdram_init

    /* Set the stack pointer so that we can jump to C-code...
	 */
    ldr		r1, =(MonStack + MONSTACKSIZE - 4)
	mov     sp, r1                      

	/* call the remaining init code in C
	 */
	bl	hw_init2
	nop
	nop

/********************************************************************/

midstart:
    ldr r0, =INITIALIZE
    
    /* fall-through to 'lukewarmstart' */

/********************************************************************/
    
lukewarmstart:  
    /* Save the argument to r11 */
    mov r11, r0

    /*
     * *** DO NOT TOUCH R11 ***
     */ 

    /*
     * Set-up the stack-pointers for all operating modes
     */

    /* FIQ mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x11           /* set FIQ mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(FiqStack + FiqStackSz - 4)    /* initialize the stack ptr */
    /* IRQ mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x12           /* set IRQ mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(IrqStack + IrqStackSz - 4)    /* initialize the stack ptr */
    /* Abort mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x17           /* set Abort mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(AbtStack + AbtStackSz - 4)    /* initialize the stack ptr */
    /* Undef mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x1b           /* set Undef mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(UndStack + UndStackSz - 4)    /* initialize the stack ptr */
    /* System mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x1f           /* set System mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(SysStack + SysStackSz - 4)    /* initialize the stack ptr */
    /* 'warmstart' will take us back to SVC mode
       stack for SVC mode will also be setup in warmstart */

    mov r0, r11     /* get argument back from r11 */
    b   warmstart

    
/********************************************************************/
    
warmstart:
    /* Save the argument to r11 */
    mov r11, r0

    /*
     * *** DO NOT TOUCH R11 ***
     */ 


    /* Change (back) to SVC mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x13           /* set System mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    /* Reset the stack pointer for the SVC mode (our current mode) */
    ldr sp, =(MonStack + MONSTACKSIZE - 4)

    /*
     * Restore argument which was saved to r11 and jump to
     * the C function start().
     */ 
    
    mov r0, r11
jump_to_c:
    bl start

    /* the C code should never return */
    b coldstart

.align 4


/*********************************************************************
 *
 * simple delay loop
 */
delay_5000:
	ldr		r3, =5000			/* loop count */
delay_loop:
	subs	r3,r3,#1
	bne		delay_loop
	nop

	mov		pc, lr

/*********************************************************************
 *
 * SDRAM Initialization
 */
sdram_init:
	ldr	r1, =0x78180014 /* DRAM_RFSH0, use value in RFRSH1 x 1 */
	ldr	r0, =0x00000001
	str	r0, [r1]

	ldr	r1, =0x7818001c /* DRAM_RFSH1, CCLK/64000 - assume 33Mhz worst case */
	ldr	r0, =0x000001d6
	str	r0, [r1]

	ldr	r1, =0x78180000 /* DRAM_BWC, bus width = 16 bits */
	ldr	r0, =0x00000002
	str	r0, [r1]

	ldr	r1, =0x78180008 /* DRAM_RPC, access paramters = DRAM_RPC_SD_2_4_2_2 */
	ldr	r0, =0x00000004
	str	r0, [r1]

	ldr	r1, =0x78180010 /* DRAM_CMD, send 2 nop commands */
	ldr	r0, =0x00000000
	str	r0, [r1]
	str	r0, [r1]

	ldr	r1, =0x78180010 /* DRAM_CMD, send a pre-charge all command */
	ldr	r0, =0x00000004
	str	r0, [r1]

	ldr	r1, =0x78180010 /* DRAM_CMD, send 8 refresh commands */
	ldr	r0, =0x00000005
	str	r0, [r1]
	str	r0, [r1]
	str	r0, [r1]
	str	r0, [r1]
	str	r0, [r1]
	str	r0, [r1]
	str	r0, [r1]
	str	r0, [r1]

	ldr	r1, =0x78180004	/* 64MB SDRAM_RMC, 8 column bits, SYNC, CL2, CBR */
	ldr	r0, =0x00000080
	str	r0, [r1]

	ldr	r1, =0x7818000c	/* DRAM_SDMD, set CL2 in SDRAM mode register */
	ldr	r0, =0x00000080
	str	r0, [r1]

	ldr	r1, =0x78180018 /* DRAM_RDWC, set power down delay */
	ldr	r0, =0x00000010 /* (only valid if power down is enabled - we don't) */
	str	r0, [r1]

	mov		pc, lr

