//==========================================================================
//
//      mc9328mx21_lcd.c
//
// Author(s):   Michael Kelly - Cogent Computer Systems, Inc.
// Date:        05-27-2002
// Description:	Init Code for MC9328mx21 LCD Controller
// NOTE 		Only 16-bit mode has been tested!
//
//==========================================================================

#include "config.h"
#include "cpuio.h"
#include "stddefs.h"
#include "genlib.h"
#include "mc9328mx21.h"
#include "mc9328mx21_lcd.h"
#include "cpu_gpio.h"
#include "vga_lookup.h"
#include "fb_draw.h"

//--------------------------------------------------------------------------
// function prototypes and externs
//
void lcd_init(void);
void lcd_on(void);
void lcd_off(void);

extern void udelay(int);
extern int GPIO_clr(int);
extern int GPIO_set(int);
extern int GPIO_tst(int);
extern int GPIO_out(int);
extern int GPIO_in(int);

// globals to keep track of foreground, background colors and x,y position
int lcd_color_depth;		// 4, 8 or 16
int lcd_fg_color;			// 0 to 15, used as lookup into VGA color table
int lcd_bg_color;			// 0 to 15, used as lookup into VGA color table
ulong lcd_fb_offset;	   	// current offset into frame buffer for lcd_putchar

//--------------------------------------------------------------------------
// lcd_init
//
// This function sets up the MC9328MXL LCD controller
//
void
lcd_init(void)
{
	static int beenhere;
	ushort temp16;
#ifdef VGA_MOD			// See comments at top of vga_mod_cpuio.h
	long int x;
#endif
	int i;

	if (!beenhere) {
		printf("MC9328MX21 LCD Initialization.\n");
		beenhere = 1;
	}

#ifdef VGA_MOD
	// Change perdev3 to 4
	// to make PERCLK3 = 53.2 mhz (266 / 5)
	x = PLL_REG(PLL_PCDR1);
	x &= 0xff00ffff;
	x |= 0x00040000;
	PLL_REG(PLL_PCDR1) = x;
#endif

	lcd_color_depth = 16;

    // From sample code from Motorola
    //
    LCDC_REG(LCDC_SSA) = LCD_BUF_ADD;

	// Note, LCD pixel clock is derived from PERCLK2 which
	// defaults to 96Mhz.
    // Panel Control Register

    LCDC_REG(LCDC_PCR) = LCDC_PCR_TFT
                       | LCDC_PCR_COLOR
                       | LCDC_PCR_PIXPOL_HI
                       | LCDC_PCR_FLMPOL_LO
                       | LCDC_PCR_LPPOL_LO
                       | LCDC_PCR_CLKPOL_TFT_HI
                       | LCDC_PCR_OEPOL_HI
                       | LCDC_PCR_END_LITTLE
					   | LCDC_PCR_PBSIZ8			// forced to 16 for TFT
					   | LCDC_PCR_BPIX16			// 16-bit/pixel only
                       | LCDC_PCR_SWAP_SCLKSEL
                       | LCDC_PCR_SWAP_PCD(LCD_PCD);   // 16Mhz/(PCD+1)

    //  LCDC_HCR - Horizontal Configuration Register
    LCDC_REG(LCDC_HCR) = LCDC_HCR_H_WIDTH(LCD_H_WIDTH)     // Pulse Width in pixels
                       | LCDC_HCR_H_WAIT_1(LCD_H_FRONT)    // Front Porch
                       | LCDC_HCR_H_WAIT_2(LCD_H_BACK);   // Back Porch
             
    //  LCDC_VCR - Vertical Configuration Register
    LCDC_REG(LCDC_VCR) = LCDC_VCR_V_WIDTH(LCD_V_WIDTH)     // Pulse Width in pixels
                       | LCDC_VCR_V_WAIT_1(LCD_V_FRONT)    // Front Porch
                       | LCDC_VCR_V_WAIT_2(LCD_V_BACK);   // Back Porch


    LCDC_REG(LCDC_SIZE) = LCDC_SIZE_XMAX(PIXELS_PER_ROW/16)	// lower 4 bits ignored, so must be on 16 byte boundry
                        | LCDC_SIZE_YMAX(PIXELS_PER_COL);
    
	LCDC_REG(LCDC_POS) = 0; // left/right panning register offset

    LCDC_REG(LCDC_VPW) = (PIXELS_PER_ROW/2);
	
    LCDC_REG(LCDC_SSA) = LCD_BUF_ADD;

    LCDC_REG(LCDC_DMACR) = 0x80080004;

	// fill the screen with the background color
	switch (lcd_color_depth){
		case 4:
			temp16 = lcd_bg_color | lcd_bg_color << 4 | lcd_bg_color << 8 | lcd_bg_color << 12;
			break;
		case 8:
			temp16 = lcd_bg_color | lcd_bg_color << 8;
			break;
		default:
			temp16 = lcd_bg_color;
			break;
	}
	for (i = 0; i < (PIXELS_PER_COL * PIXELS_PER_ROW) * 4; i += 2){	
		LCD_BUF(i) = temp16;
	}

	return;
}

//--------------------------------------------------------------------------
// lcd_on
//
// This function turns on the MC9328MX21 LCD Controller and the Backlight Inverter
//
void lcd_on()
{
	// Make sure it's off first
//	lcd_off();

	// Enable the clock to the LCDC
	PLL_REG(PLL_PCCR0) |= BIT26; // bit 26 controls the LCD clock

	// power up the LCD
    udelay(10000);
    GPIO_set(LCD_VEE);
	udelay(10000);
    GPIO_set(LCD_VEN);
	udelay(10000);
    GPIO_set(LCD_BKL);
}

//--------------------------------------------------------------------------
// lcd_off
//
// This function turns off the LCD Controller and the LCD backlight
//
void lcd_off()
{
	// power down the LCD
    udelay(10000);
    GPIO_clr(LCD_BKL);
	udelay(10000);
    GPIO_clr(LCD_VEN);
	udelay(10000);
    GPIO_clr(LCD_VEE);
	
	// Disable the clock to the LCDC
	PLL_REG(PLL_PCCR0) &= ~BIT26; // bit 26 controls the LCD clock


}

char *lcd_tstHelp[] = {
    "MC9328MX21 LCD controller test",
    "-[n,x,d[4,8,16]]",
	"The user may set color depth to run the test at.",
	"The frame buffer R/W test will test all of the frame ",
	"buffer regardless of depth.",
    "Options...",
    "  -n    run test without keycheck - CAUTION: RESET SYSTEM TO STOP!",
	"  -d4   run test, force a depth of 4-bits/pixel",
	"  -d8   run test, force a depth of 8-bits/pixel",
	"  -d16  run test, force a depth of 16-bits/pixel",
	"  -x    init only, do not run frame buffer tests",
	"",
	"  No options, default to current mode and depth.",
	0
};

int lcd_tst(int argc,char *argv[])
{
	volatile ushort wr16, rd16;
	int i, x, opt;
	int no_wait = 0;
	int init_only = 0;
	char c;

	lcd_off();

    while ((opt=getopt(argc,argv,"clnsxd:4,8,16")) != -1) {
        switch(opt) {
			case 'd':	// set the color depth
                switch(*optarg) {
            	    case '4':
	        	        lcd_color_depth = 4;
						printf("Forcing 4bpp Mode!\n");
        	            break;
    	            case '8':
            	        lcd_color_depth = 8;
						printf("Forcing 8bpp Mode!\n");
                    	break;
	                default:	// test with 16bpp
	                    lcd_color_depth = 16;
						printf("Forcing 16bpp Mode!\n");
						break;
                }
	            break;
	        case 'n':   // no waiting for keypress - fastest operation
				no_wait = 1;
				printf("No Keypress Mode, Must Reset System to Stop!\n");
				break;
	        case 'x':   // init only 
				no_wait = 1;
				printf("Initializing LCD, Skipping testsp!\n");
				init_only = 1;
				break;
			default:	// test with current mode
				break;
		}
	}

	// get the new parameters into the LCD controller
	lcd_init();
	lcd_on();

	if (init_only) return 0;

	printf("Frame Buffer R/W...");
	// do an address=data read/write test on the frame buffer
	// PIXELS_PER_COL * PIXELS_PER_ROW is the highest pixel.  
	// Multiply by bits_per_pixel (sed_color_depth), then
	// divide by 8 to get the actual byte count.
	for (i = LCD_ROW_SIZE(lcd_color_depth); i < LCD_FB_SIZE(lcd_color_depth) + LCD_ROW_SIZE(lcd_color_depth); i += 2){	
		LCD_BUF(i) = i & 0xffff;
		rd16 = LCD_BUF(i);
		if(rd16 != (i & 0xffff)){
			printf("Fail at 0x%08x, WR 0x%08x, RD 0x%04lx!\n",LCD_BUF_ADD + i, i, (ulong)rd16);
			return -1;
		}
	} 

	printf("OK!, Press key to continue.\n");

	c = getchar();

	printf("Frame Buffer Start: 0x%08x, End 0x%08x\n",LCD_BUF_ADD, 
			LCD_BUF_ADD + LCD_FB_SIZE(lcd_color_depth) + LCD_ROW_SIZE(lcd_color_depth));
	if (no_wait)
	{
		printf("Begin Full Screen Color Test.\n");
		while(1){
			// fill the frame buffer with incrementing color values
			for (x = 0; x < 16; x++){
				switch (lcd_color_depth){
					case 4:  wr16 = x | x << 4 | x << 8 | x << 12; break;
					case 8:  wr16 = x | x << 8; break;
					default: wr16 = vga_lookup[x]; break;	// 16-bits bypasses the lookup table
				}
				for (i = 0; i < LCD_FB_SIZE(lcd_color_depth); i += 2){	
					LCD_BUF(i) = wr16;
				}
			} // for x
		} // while
	} // no_wait
	else
	{
		printf("Begin Full Screen Color Test, Press any key to go to next color, \'x\' to end.\n");
		while(1){
			// fill the frame buffer with incrementing color values
			for (x = 0; x < 16; x++){
				switch (lcd_color_depth){
					case 4:  wr16 = x | x << 4 | x << 8 | x << 12; break;
					case 8:  wr16 = x | x << 8; break;
					default: wr16 = vga_lookup[x]; break;	// 16-bits bypasses the lookup table
				}
				for (i = LCD_ROW_SIZE(lcd_color_depth); i < LCD_FB_SIZE(lcd_color_depth) + LCD_ROW_SIZE(lcd_color_depth); i += 2){	
					LCD_BUF(i) = wr16;
				}
				c = getchar();
				if (c == 'x') goto lcd_tst_next;
			} // for x
		} // while
	} // else no keycheck test

	lcd_tst_next:

	// write a one pixel border around the screen
	lcd_fg_color = 15;	// VGA Bright White
	fb_draw_line2(LEFT, TOP, RIGHT, TOP, lcd_fg_color);
	fb_draw_line2(LEFT, TOP, LEFT, BOTTOM, lcd_fg_color);
	fb_draw_line2(LEFT, BOTTOM, RIGHT, BOTTOM, lcd_fg_color);	// bottom left to bottom right
	fb_draw_line2(RIGHT, TOP, RIGHT, BOTTOM, lcd_fg_color);	// bottom right to top right
	// draw an x
	fb_draw_line2(LEFT, TOP, RIGHT, BOTTOM, lcd_fg_color);
	fb_draw_line2(LEFT, BOTTOM, RIGHT, TOP, lcd_fg_color);
	// draw 3 circles at the center of the screen, one inside the other
	lcd_fg_color = 12;	// VGA Bright Red
	fb_draw_circle2(CENTER_X, CENTER_Y, 100, lcd_fg_color);
	lcd_fg_color = 10;	// VGA Bright Green
	fb_draw_circle2(CENTER_X, CENTER_Y, 66, lcd_fg_color);
	lcd_fg_color = 9;	// VGA Bright Blue
	fb_draw_circle2(CENTER_X, CENTER_Y, 33, lcd_fg_color);

	return 0;
}

// fb_set_pixel sets a pixel to the specified color.
// This is target specific and is called from the generic
// fb_draw.c functions
void fb_set_pixel(ulong X, ulong Y, uchar color)
{
   	// Make sure the specified pixel is valid.
#if 0
   	if((X < 0) || (X >= PIXELS_PER_ROW) || (Y < 0) || (Y >= PIXELS_PER_COL))
   	{   
		printf("fb_set_pixel() bad X (%ld) or Y (%ld)!\n", X, Y);
		return;
   	}
#else
	if (X < 0)
		X = 0;
	else {
		if (X >= PIXELS_PER_ROW)
			X = PIXELS_PER_ROW-1;
	}
	if (Y < 0)
		Y = 0;
	else {
		if (Y >= PIXELS_PER_COL)
			Y = PIXELS_PER_COL-1;
	}
#endif

	LCD_BUF(LCD_GET_PIXEL_ADD(X, Y)) = vga_lookup[color];
}

