/* fat.c
    Author: David de la Cruz, Technology Rescue.
    email:  ddlc@techresc.net

	This source file defines FAT Filesystem routines.

	Modification History :

	Version 2.1:
		29.04.2004 : new.
		06.05.2004 : clean-up.
*/

#include "config.h"
#if INCLUDE_FAT

#include "genlib.h"
#include "stddefs.h"
#include "errno.h"
#include "fs.h"
#include "fat.h"

/* Local variables. */
file_system_ops * current_media = NULL;

uchar dir_sector[FS_BLOCK_SIZE];	// Used to contain current root directory sector.
uchar fat_sector[FS_BLOCK_SIZE];	// Used to contain current fat sector.
int	fat_was_modified;				// Flag to know if fat_sector was modified and should be saved.
uchar saved_fat_sec[FS_BLOCK_SIZE];	// Used to contain previous fat sector.
int fat_miss;						// Flag to signal that a cluster number was not saved in previous fat sector.
ulong saved_fat_entry;				// Used to keep fat entry address corresponding to saved_fat_sec.


/* init_media function.

	Entry :
		file_system_ops * fsops : filled in with media functions.

	Exit :
		return 0 if successfull, else error code.

	Function :
		It initialises list_of_media.
		This function is called by media module such as mmc.

*/
int init_media(file_system_ops * fsops)
{
#ifdef FAT_DEBUG
	printf("init_media.\n");
#endif

	current_media = fsops;

	return(0);
}

/* init_sector0 function.

	Entry :
		master_boot_record * MBR : master boot record initialised with 0s.
		ulong * sector0 : uninitialised buffer.

	Exit :
		master_boot_record * MBR : master boot record initialised with first sector.
		ulong * sector0 : buffer which contains sector 0 datas.
		return 0 if successfull, else error code.

	Notes :
		MBR and sector0 may contain same data if media is floppy-like.
*/
int init_sector0(master_boot_record * MBR, ulong * sector0)
{
	int	ret_val = -1;	// Assume failure.
	int part_count = 0;
	int	mbr_found = 0;

#ifdef FAT_DEBUG
	printf("init_sector0.\n");
#endif

	// Read first sector. Assume it is boot sector and store it into sector0.
	current_media->media_read((uchar *)sector0, current_media->start_address, FS_BLOCK_SIZE);

	// Copy this sector to MBR structure.
	memcpy((char *)MBR, (char *)sector0, sizeof(master_boot_record));

	// Check for MBR.
	if(MBR->signature == 0xAA55)
	{
#ifdef FAT_DEBUG
		printf("Finding partitions or boot sector.\n");
#endif
		ret_val = 0;	// This sector is either MBR or boot sector.
		mbr_found = 0;
		for(part_count=0; part_count < 4; part_count++)
		{
			if((MBR->Part[part_count].boot_ind == BOOTABLE) || (MBR->Part[part_count].boot_ind == NBOOTABLE))
			{
				if((MBR->Part[part_count].sys_ind == PDOS16) || (MBR->Part[part_count].sys_ind == PDOS32))
				{	// This sector is MBR.
					mbr_found = 1;
					break;
				}
			}
		}
	}
	if(mbr_found)
	{	// Read boot sector according to partition infos.
#ifdef FAT_DEBUG
		printf("MBR found.\n");
		print_mbr(MBR);
#endif
		current_media->start_address = CFG_MMC_BASE + (MBR->Part[part_count].start * FS_BLOCK_SIZE);
		current_media->media_read((uchar *)sector0, current_media->start_address, FS_BLOCK_SIZE);
	}
#ifdef FAT_DEBUG
	else if(ret_val == 0)
	{
		printf("boot sector found.\n");
	}
#endif

	return ret_val;
}

/* init_boot_sector function.

	Entry :
		boot_sector * BPB : boot sector structure initialised with 0s.
		ulong * sector0 : buffer which contains sector 0 datas.

	Exit :
		boot_sector * BPB : boot sector structure filled in with sector0 area.
		return 0 if successfull, else error code.
*/
int init_boot_sector(boot_sector * BPB, ulong * sector0)
{
#ifdef FAT_DEBUG
	printf("init_boot_sector.\n");
#endif

	// Copy sector0 data to BPB.
	memcpy((char *)BPB, (char *)sector0, sizeof(boot_sector));

#ifdef FAT_DEBUG
	print_boot_sector(BPB);
#endif

	return 0;
}

/* init_file_system function.

	Entry :
		file_system * fs : file system structure initialised with 0s.
		boot_sector * BPB : boot sector structure initialised by init_boot_sector function.

	Exit :
		file_system * fs : file system structure filled in with boot sector infos.
		return 0 if successfull, else error code.
*/
int init_file_system(file_system * fs, boot_sector * BPB)
{
	int ret_val = -1;	// Assume failure.

#ifdef FAT_DEBUG
	printf("init_file_system.\n");
#endif

	// cluster size in bytes.
	fs->cluster_size = BPB->sector_size * BPB->cluster_size;

	// # sectors in root directory.
	fs->root_dir = ((BPB->dir_entries * 32) + (BPB->sector_size-1)) / BPB->sector_size;

	// Determine the count of sectors in the FAT.
	if(BPB->fat_length != 0)
		fs->fat_size = BPB->fat_length;
	else
		fs->fat_size = BPB->ext.fat32.fat32_length;

	// Determine the count of sectors in the media.
	if(BPB->sectors != 0)
		fs->tot_sec = BPB->sectors;
	else
		fs->tot_sec = BPB->total_sect;

	// Determine the count of sectors in the data region.
	fs->data_sec = fs->tot_sec - (BPB->reserved + (BPB->fats * fs->fat_size) + fs->root_dir);

	// Determine the count of clusters.
	fs->count_of_clus = fs->data_sec / BPB->cluster_size;

	// Check FAT type validity.
	if(fs->count_of_clus < 4085)
	{	// FAT12, unsupported.
		fs->fat_type = 0;
		ret_val = -ENOSYS;	// Indicate function not implemented error code.
	}
	else if(fs->count_of_clus < 65525)
			{	// FAT16.
				fs->fat_type = PDOS16;
				ret_val = 0;	// Indicate success.
			}
			else
			{
				fs->fat_type = PDOS32;
				ret_val = 0;	// Indicate success.
			}

	if(fs->fat_type != 0)
	{	// Compute offsets.
		fs->boot_sector_off = current_media->start_address;
		fs->fat_off = fs->boot_sector_off + (BPB->reserved * BPB->sector_size);
		if(BPB->fats > 1)
			fs->fat_save_off = fs->fat_off + (fs->fat_size * BPB->sector_size);
		else
			fs->fat_save_off = 0;
		fs->root_dir_off = fs->fat_off + ((BPB->fats * fs->fat_size) * BPB->sector_size);
		fs->data_off = fs->root_dir_off + (fs->root_dir * BPB->sector_size);

		fs->BPB = BPB;
	}

#ifdef FAT_DEBUG
	print_file_system(fs);
#endif
	return ret_val;
}

#if defined(FAT_DEBUG) || INCLUDE_FS

/* print_file_system function.

	Entry :
		file_system * fs : file system structure to print.

	Exit :
		no return
*/
void print_file_system(file_system * fs)
{
	printf("Boot sector offset   = %lx\n", fs->boot_sector_off);/* Boot sector offset related to partition start. */
	printf("FAT Offset           = %lx\n", fs->fat_off);		/* FAT sector offset related to partition start. */
	printf("Saved FAT Offset     = %lx\n", fs->fat_save_off);	/* Saved FAT sector. */
	printf("Root dir offset      = %lx\n", fs->root_dir_off);	/* Root directory offset. */
	printf("First cluster offset = %lx\n", fs->data_off);		/* First Cluster offset. */

	printf("# sec/root           = %lx\n", fs->root_dir);		/* # sectors in root directory. */
	printf("# sec/FAT            = %lx\n", fs->fat_size);		/* # sector per FAT. */
	printf("# total sec          = %lx\n", fs->tot_sec);		/* # total sectors */
	printf("# data sec           = %lx\n", fs->data_sec);		/* # data sectors. */
	printf("Count of Clusters    = %lx\n", fs->count_of_clus);	/* Count of clusters. This field is used for fat type determination. */
	printf("FAT Type             = %lx\n", fs->fat_type);		/* Keep fat type.*/
}

#endif

#ifdef FAT_DEBUG
/* print_mbr function.

	Entry :
		master_boot_record * mbr : file system structure to print.

	Exit :
		no return
*/
void print_mbr(master_boot_record * mbr)
{
	int count;

	for(count = 0; count <4; count++)
	{
		printf("Part %d :\n", count);
		printf("%lx ",mbr->Part[count].boot_ind);		/* 0x80 = boot device - 0x00 = non-bootable */
		printf("%lx ",mbr->Part[count].head);			/* starting head */
		printf("%lx ",mbr->Part[count].sector);			/* starting sector */
		printf("%lx ",mbr->Part[count].cyl);			/* starting cylinder */
		printf("%lx ",mbr->Part[count].sys_ind);		/* What partition type */
		printf("%lx ",mbr->Part[count].end_head);		/* end head	*/
		printf("%lx ",mbr->Part[count].end_sector);		/* end sector */
		printf("%lx ",mbr->Part[count].end_cyl);		/* end cylinder	*/
		printf("%lx ",mbr->Part[count].start);			/* starting sector counting from 0 */
		printf("%lx ",mbr->Part[count].size);			/* nr of sectors in partition */
		printf("\n");

	}
}

/* print_boot_sector function.

	Entry :
		boot_sector * bs : boot sector structure to print.

	Exit :
		no return
*/
void print_boot_sector(boot_sector * bs)
{
	printf("boot sector infos : \n");
	printf("%lx ",bs->sector_size);		/* Bytes/sector */
	printf("%lx ",bs->cluster_size);	/* Sectors/cluster */
	printf("%lx ",bs->reserved);		/* Number of reserved sectors */
	printf("%lx ",bs->fats);			/* Number of FATs */
	printf("%lx ",bs->dir_entries);		/* Number of root directory entries */
	printf("%lx ",bs->sectors);			/* Number of sectors */
	printf("%lx ",bs->media);			/* Media code */
	printf("%lx ",bs->fat_length);		/* Sectors/FAT */
	printf("%lx ",bs->secs_track);		/* Sectors/track */
	printf("%lx ",bs->heads);			/* Number of heads */
	printf("%lx ",bs->hidden);			/* Number of hidden sectors */
	printf("%lx ",bs->total_sect);		/* Number of sectors (if sectors == 0) */
	printf("\n");
}

/* print_file_infos function.

	Entry :
		file_infos * fi : file infos structure to print.

	Exit :
		no return
*/
void print_file_infos(file_infos * fi)
{
	int count;

	printf("Filename = ");
	for(count=0; count<(FILENAME_LEN+FILENAME_EXT); count++)
	{
		printf("%c", fi->filename[count]);
	}

	printf("\nDirectory entry addr   = %lx", fi->dir_entry_addr);		/* Directory entry address. */
	printf("\nCluster #              = %lx", fi->cluster_num);	/* First cluster of cluster chain. */
	printf("\nfile size              = %lx", fi->file_size);			/* File size in bytes. */
	printf("\nCurrent FAT entry addr = %lx", fi->current_fat_entry);	/* Address of fat entry for the current cluster. */
	printf("\nCurrent cluster addr   = %lx\n", fi->current_cluster);	/* Current cluster address. */

}

#endif

/* find_filename function.

	Entry :
		const char * filename : filename string to search.
		const file_system * fs : file system structure.

	Exit :
		file_infos structure filled in with directory entry content.
		file_infos->cluster_num = 0 if file not found.
*/
file_infos find_filename(const char * filename, const file_system * fs)
{
	file_infos retval = {"", 0, 0, 0, 0, 0};
	int count, count_ext;
	int end_of_filename = 0;	// Indicate end of filename is reached.
	int	dot_found = 0;			// Indicate extention.
	int dir_entry_found = 0;	// Indicate file entry is found in root directory.
	dir_entry *dir;

	// First, compute file name (8+.+3) to directory name format (11).
	for(count = 0; count<(FILENAME_LEN+FILENAME_EXT); count++)
	{	// init each char to ' ';
		retval.filename[count] = ' ';
	}
	retval.filename[count] = '\0';	// Last char to NULL.

	// Copy the name part of the 8.3 name.
	for(count = 0; count < FILENAME_LEN; count++)
  	{
		if(filename[count] == '.')
		{
			dot_found = 1;		// Indicate dot is found in filename.
			count++;			// Increment count to exclude dot char.
			break;
		}
		if(filename[count] == '\0')
		{
			end_of_filename = 1;// Indicate end of filename.
			break;
		}
		retval.filename[count] = filename[count];

		/* uppercase the name if it's a lower case alpha
		*/
		if((retval.filename[count]  >= 'a') && (retval.filename[count]  <= 'z'))
		{
			retval.filename[count] &= ~('a' - 'A');
		}
  	}

	// Check if we've got extention.
	if((count==FILENAME_LEN) && filename[count] == '.')
	{
		dot_found = 1;		// Indicate dot is found in filename.
		count++;			// Increment count to exclude dot char.
	}
	

	// Now, copy extention part of the 8.3 name.
	if(dot_found && !end_of_filename)
	{
		for(count_ext=FILENAME_LEN; count_ext < (FILENAME_LEN+FILENAME_EXT); count++, count_ext++)
		{
			if(filename[count] == '\0')
			{
				end_of_filename = 1;// Indicate end of filename.
				break;
			}
			retval.filename[count_ext] = filename[count];
	
			/* uppercase the name if it's a lower case alpha
			*/
			if((retval.filename[count_ext] >= 'a') && (retval.filename[count_ext] <= 'z'))
			{
				retval.filename[count_ext] &= ~('a' - 'A');
			}
		}
	}

	// Find retval.filename in root directory.
	retval.dir_entry_addr = fs->root_dir_off;

	// Read directory sector from media and init dir_entry structure.
	current_media->media_read((uchar *)dir_sector, retval.dir_entry_addr, fs->BPB->sector_size);
	dir = (dir_entry *)dir_sector;	// dir points to first directory entry.

	while(!dir_entry_found)
	{	// Exit when filename is found in root directory.
		// This loops will be break when end of directory is reached.

		// First, check first directory entry validity.
		if(dir->name[0] == 0x0)
		{
			// This value indicate there's no more directory entry.
			break;
		}

		if(strncmp(dir->name, retval.filename, FILENAME_LEN+FILENAME_EXT))
		{	// filenames are not the same; go next dir entry.
			// Check validity of next dir address.
			if((ulong *)(dir+1) > (ulong *)(dir_sector+FS_BLOCK_SIZE))
			{	// Need to load next directory sector.
				retval.dir_entry_addr+=FS_BLOCK_SIZE;
				current_media->media_read((uchar *)dir_sector, retval.dir_entry_addr, fs->BPB->sector_size);
			}
			else
			{	// If valid, increment dir pointer.
				dir++;
			}
		}
		else
		{
			dir_entry_found = 1;		// Indicate file name is found into directory entry.
		}
	}

	if(dir_entry_found)
	{	// filename was found into directory sector.
		retval.dir_entry_addr = (ulong)dir - (ulong)dir_sector + fs->root_dir_off;
		retval.cluster_num = dir->start | dir->starthi << 16;
		retval.file_size = dir->size;
		retval.current_cluster = fs->data_off + (retval.cluster_num - 2)*(fs->cluster_size);
		if(fs->fat_type == PDOS32)
			retval.current_fat_entry = fs->fat_off + 4*retval.cluster_num;
		else	// FAT16.
			retval.current_fat_entry = fs->fat_off + 2*retval.cluster_num;

		// Read FAT sector where this cluster entry is.
		current_media->media_read(fat_sector, retval.current_fat_entry & ~(FS_BLOCK_SIZE - 1), FS_BLOCK_SIZE);
	}
	else
	{	// filename was not found, indicate failure with cluster_num = 0.
		retval.cluster_num = 0;
		// keep this free directory entry.
		retval.dir_entry_addr = (ulong)dir - (ulong)dir_sector + fs->root_dir_off;
		// First FAT sector to search for free entry.
		retval.current_fat_entry = fs->fat_off;
		// Read this first FAT sector.
		current_media->media_read(fat_sector, retval.current_fat_entry & ~(FS_BLOCK_SIZE - 1), FS_BLOCK_SIZE);
	}
	fat_was_modified = 0;
	fat_miss = 0;

	return retval;
}

/* read_cluster function.

	Entry :
		const char * filename : filename string to search.
		const file_system * fs : file system structure.
		ulong dest : destination address to write to.

		<0 : error in file read.
		>0 : bytes read from file and written to dest address.
*/
int read_cluster(const file_infos * fi, const file_system * fs, ulong dest)
{
	int retval = -1;	// Assume error.
	ulong src;			// Source address to read from.
	ulong sizetoread;	// Size of cluster.

	sizetoread = fs->cluster_size;
	src = fi->current_cluster;

#ifdef FAT_DEBUG
	printf("%lx to read from %lx at dest %lx\n", sizetoread, src, (uchar *)dest);
#endif

	retval = current_media->media_read((uchar *)dest, src, sizetoread);

	return retval;
}

/* find_next_cluster function.

	Entry :
		const file_infos * fi : contains current file infos.
		const file_system * fs : file system structure.

	Exit :
		file_infos structure filled in with fi updated = next cluster infos.
		file_infos->cluster_num = 0 if file not found.
*/
file_infos find_next_cluster(const file_infos * fi, const file_system * fs)
{
	static file_infos retval;
	int cur_fat_sector_number;			// To load fat sector from media.
	int next_fat_sector_number;			// To load fat sector from media.
	ulong cluster_offset;

	// First, copy fi to retval.
	retval = *fi;

	cluster_offset = fi->current_fat_entry - fs->fat_off;
	if(fs->fat_type == PDOS32)
	{	// Read next cluster number from current FAT entry in fat_sector table.
		retval.cluster_num = *(ulong *)((cluster_offset & (FS_BLOCK_SIZE - 1) ) + fat_sector);

		// Compute FAT entry address for next cluster.
		retval.current_fat_entry = fs->fat_off + 4*retval.cluster_num;
	}
	else
	{	// Read next cluster number from current FAT entry.
		retval.cluster_num = *(ushort *)((cluster_offset & (FS_BLOCK_SIZE - 1) ) + fat_sector);

		// Compute FAT entry address for next cluster.
		retval.current_fat_entry = fs->fat_off + 2*retval.cluster_num;
	}

	// Compute next cluster address.
	retval.current_cluster = fs->data_off + (retval.cluster_num - 2)*(fs->cluster_size);

	// Check fat_sector validity if not end of chain.
	if(! (((fs->fat_type == PDOS32) && (retval.cluster_num == PDOS32_EOC))
	|| ((fs->fat_type == PDOS16) && (retval.cluster_num == PDOS16_EOC))) )
	{
		cur_fat_sector_number = (fi->current_fat_entry & ~(FS_BLOCK_SIZE - 1)) - fs->fat_off;
		next_fat_sector_number = (retval.current_fat_entry & ~(FS_BLOCK_SIZE - 1)) - fs->fat_off;
	
		if(cur_fat_sector_number != next_fat_sector_number)
		{	// fat_sector need to be updated from media.
#ifdef FAT_DEBUG
			printf("Read FAT from %lx at dest %lx\n", retval.current_fat_entry & ~(FS_BLOCK_SIZE - 1), fat_sector);
#endif
			current_media->media_read(fat_sector, retval.current_fat_entry & ~(FS_BLOCK_SIZE - 1), FS_BLOCK_SIZE);
			fat_was_modified = 0;
		}
	}
	
	return retval;	// Return next cluster file infos.
}

/* init_dir_entry function.

	Entry :
		file_infos * fi : contains current file infos.
		const file_system * fs : file system structure.

	Exit :
		file_infos * fi : contains updated file infos such as file name

*/
void init_dir_entry(file_infos * fi, const file_system * fs)
{
	dir_entry *dir;
	int count;

	// Create Directory entry in dir_sector.
	dir = (dir_entry*)(fi->dir_entry_addr - fs->root_dir_off + (ulong)dir_sector);
	for(count = 0; count<(FILENAME_LEN+FILENAME_EXT); count++)
	{	// Fill in filename field.
		dir->name[count] = fi->filename[count];
	}
#ifdef FAT_DEBUG
	printf("dir address = %lx\n", dir_sector);
#endif
}

/* find_free_cluster function.

	Entry :
		const file_infos * fi : contains current file infos.
		const file_system * fs : file system structure.

	Exit :
		file_infos structure filled in with fi updated = next cluster infos.
		file_infos->cluster_num = 0 if file not found.
*/
file_infos find_free_cluster(const file_infos * fi, const file_system * fs)
{
	static file_infos retval;
	ulong * fat32_entry;		// Pointer to index fat_sector table (FAT32).
	ushort * fat16_entry;		// Pointer to index fat_sector table (FAT16).
	ulong fat_sector_content;

	// Flag to set when free cluster is found.
	int free_cluster_found;

	// Var to browse fat_sector table.
	static int count = 0;

	// This variable keep last free cluster entry.
	static ulong current_free_cluster = 0;	// First search start at offset 0.

#ifdef FAT_DEBUG
	printf("Find free cluster.\n");
#endif

	// First, copy fi to retval.
	retval = *fi;

	// At this point, it is assume that fat_sector is valid.
	// Then, search for free cluster in fat_sector table.
	free_cluster_found = 0;					// Free cluster is not found yet.
	count = current_free_cluster;			// Init count with saved current_free_cluster.
	fat32_entry = (ulong *)(fat_sector+count);		// Init FAT32 index pointer.
	fat16_entry = (ushort *)(fat_sector+count);		// Init FAT16 index pointer.
	while(!free_cluster_found)
	{
		while(count < FS_BLOCK_SIZE)
		{	// Exit when end of FAT is reached.
			if(fs->fat_type == PDOS32)
			{	// Deal with FAT32.
				fat_sector_content = (ulong)(*fat32_entry);
				if(fat_sector_content == 0x0)
				{	// free sector is found. Keep offset to current_free_cluster.
					current_free_cluster = count;
					free_cluster_found = 1;			// Indicate free cluster is found.
					*fat32_entry = PDOS32_EOC;		// Mark this entry as EOC. It will be replaced when next cluster will be written.
					fat_was_modified = 1;
					break;							// Exit.
				}
				fat32_entry++;						// Next index.
				count+=4;
			}
			else
			{	// Deal with FAT16.
				fat_sector_content = (ulong)(*fat16_entry);
				if(fat_sector_content == 0x0)
				{	// free sector is found. Keep offset to current_free_cluster.
					current_free_cluster = count;
					free_cluster_found = 1;			// Indicate free cluster is found.
					*fat16_entry = PDOS16_EOC;		// Mark this entry as EOC. It will be replaced when next cluster will be written.
					fat_was_modified = 1;
					break;							// Exit.
				}
				fat16_entry++;						// Next index.
				count+=2;
			}
		}
		if(!free_cluster_found)
		{	// Check fat_was_modified to know if fat_sector need to be saved to media.
			if(fat_was_modified)
			{
				current_media->media_write(retval.current_fat_entry & ~(FS_BLOCK_SIZE - 1), (uchar *)fat_sector, FS_BLOCK_SIZE);
#ifdef FAT_DEBUG
				printf("Fat updated at %lx.\n", retval.current_fat_entry & ~(FS_BLOCK_SIZE - 1));
#endif
				// Indicate current cluster number is not saved.
				fat_miss = 1;						
				saved_fat_entry = retval.current_fat_entry;
				// Save fat_sector to saved_fat_sec.
				memcpy(saved_fat_sec, fat_sector, FS_BLOCK_SIZE);
			}

			// Update current_fat_entry address to next.
			retval.current_fat_entry = (retval.current_fat_entry & ~(FS_BLOCK_SIZE - 1)) + FS_BLOCK_SIZE;
			// Load next fat sector.
			current_media->media_read(fat_sector, retval.current_fat_entry & ~(FS_BLOCK_SIZE - 1), FS_BLOCK_SIZE);
			fat_was_modified = 0;

#ifdef FAT_DEBUG
			printf("Read FAT from %lx at dest %lx\n", retval.current_fat_entry & ~(FS_BLOCK_SIZE - 1), fat_sector);
#endif
			// Reset count to 0.
			count = 0;
			// Reset pointers.
			if(fs->fat_type == PDOS32)
			{	// Init FAT32 index pointer.
				fat32_entry = (ulong *)fat_sector;
			}
			else
			{	// Init FAT16 index pointer.
				fat16_entry = (ushort *)fat_sector;
			}
		}
	}

	if(free_cluster_found)
	{	// Update cluster_num.
		if(fs->fat_type == PDOS32)
		{	// Deal with FAT32.
			retval.cluster_num = (count + ((retval.current_fat_entry & ~(FS_BLOCK_SIZE - 1)) - fs->fat_off)) / 4;
			retval.current_fat_entry = fs->fat_off + 4*retval.cluster_num;
		}
		else
		{	// Deal with FAT16.
			retval.cluster_num = (count + ((retval.current_fat_entry & ~(FS_BLOCK_SIZE - 1)) - fs->fat_off)) / 2;
			retval.current_fat_entry = fs->fat_off + 2*retval.cluster_num;
		}

		// Compute next cluster address.
		retval.current_cluster = fs->data_off + (retval.cluster_num - 2)*(fs->cluster_size);
#ifdef FAT_DEBUG
		printf("Free cluster %lx found at %lx (count =%lx).\n", retval.cluster_num, fat_sector, count);
#endif
	}

	return retval;

}

/* write_cluster function.

	Entry :
		const char * filename : filename string to search.
		const file_system * fs : file system structure.
		ulong src : source address to read from.

	Exit :
		<0 : error in file read.
		>0 : bytes read from file and written to dest address.
*/
int write_cluster(const file_infos * fi, const file_system * fs, ulong src)
{
	int retval = -1;	// Assume error.
	ulong dest;			// Destination address to write to.
	ulong sizetowrite;	// Size of cluster.

	sizetowrite = fs->cluster_size;
	dest = fi->current_cluster;

#ifdef FAT_DEBUG
	printf("%lx to write from %lx to %lx\n", sizetowrite, (uchar *)src, dest);
#endif

	retval = current_media->media_write(dest, (uchar *)src, sizetowrite);

	return retval;
}

/* update_fat function.

	Entry :
		file_infos * fi : current file info structure to update with next_fi->next_cluster_num.
		const file_infos * next_fi : next file info structure.
		const file_system * fs : file system structure.

	Exit :
		file_infos * fi : updated file info structure with next cluster number.
		<0 : error.
		0 : update was successfully done.
*/
int update_fat(file_infos * fi, const file_infos * next_fi, const file_system * fs)
{
	int retval = -1;		// Assume error.
	ulong * fat32_entry;	// Pointer to index fat_sector table (FAT32).
	ushort * fat16_entry;	// Pointer to index fat_sector table (FAT16).
	char * selected_fat_sec;// Used to choose between fat_sector or saved_fat_sector.

	if(fat_miss)
	{	// Used saved_fat_sec instead of fat_sector.
		selected_fat_sec = saved_fat_sec;
	}
	else
	{
		selected_fat_sec = fat_sector;
	}
	

	if(fs->fat_type == PDOS32)
	{	// Deal with FAT32.
		fat32_entry = (ulong *)((fi->current_fat_entry & (FS_BLOCK_SIZE - 1)) + selected_fat_sec);
		*fat32_entry = next_fi->cluster_num;
		fat_was_modified = 1;	// Indicate fat need to be saved.
#ifdef FAT_DEBUG
		printf("fat32 entry %lx = %lx\n", fat32_entry, *fat32_entry);
#endif
	}
	else
	{	// Deal with FAT16.
		fat16_entry = (ushort *)((fi->current_fat_entry & (FS_BLOCK_SIZE - 1)) + selected_fat_sec);
		*fat16_entry = next_fi->cluster_num;
		fat_was_modified = 1;	// Indicate fat need to be saved.
#ifdef FAT_DEBUG
		printf("fat16 entry %lx = %lx\n", fat16_entry, *fat16_entry);
#endif
	}

	// Now, check if we need to write selected_fat_sec to media or not.
	if(fat_miss)
	{	// save selected_fat_sec to media.
		retval = current_media->media_write(saved_fat_entry & ~(FS_BLOCK_SIZE - 1), (uchar *)selected_fat_sec, FS_BLOCK_SIZE);
#ifdef FAT_DEBUG
		printf("Fat updated at %lx.\n", saved_fat_entry & ~(FS_BLOCK_SIZE - 1));
#endif
		if(retval > 0)
		{	// Indicate success.
			retval = 0;
		}

		// current fat doesn't need to be written to media.
		fat_miss = 0;
		
	}
	else
	{
		retval = 0;
	}

	return retval;
}

/* update_file_system function.

	Entry :
		file_infos * fi : contains file info structure with first cluster number and file size.
		const file_system * fs : file system structure.

	Exit :
		<0 : error.
		>0 : size of file.

*/
int update_file_system(const file_infos * fi, const file_system * fs)
{
	int retval = -1;		// Assume error.
	dir_entry *dir;

#ifdef FAT_DEBUG
	printf("update_file_system %lx.\n", fat_was_modified);
#endif

	// Init dir pointer with offset from dir_sector.
	dir = (dir_entry*)(fi->dir_entry_addr - fs->root_dir_off + (ulong)dir_sector);

	// Update dir entry with file size and first cluster number.
	dir->start = fi->cluster_num & 0x0000FFFF;
	if(fs->fat_type == PDOS32)
	{	// In FAT32, cluster number are 32-bit wide.
		dir->starthi = (fi->cluster_num & 0xFFFF0000) >> 16;
	}
	dir->size = fi->file_size;

	// Update last fat sector modified.
	if(fat_was_modified)
	{
#ifdef FAT_DEBUG
		printf("Updating FAT at %lx.\n", fi->current_fat_entry & ~(FS_BLOCK_SIZE - 1));
#endif
		retval = current_media->media_write(fi->current_fat_entry & ~(FS_BLOCK_SIZE - 1), (uchar *)fat_sector, FS_BLOCK_SIZE);
	}

	// Write dir entry back to media.
#ifdef FAT_DEBUG
	printf("Updating DIR at %lx.\n", fi->dir_entry_addr & ~(FS_BLOCK_SIZE - 1));
#endif
	retval = current_media->media_write(fi->dir_entry_addr & ~(FS_BLOCK_SIZE - 1), (uchar *)dir_sector, FS_BLOCK_SIZE);
	if(retval > 0)
	{	// retval should be equal to file size.
		retval = fi->file_size;
	}

	return retval;
}


#endif /* INCLUDE_FAT */
