/* fs.c

    Author: David de la Cruz, Technology Rescue.
    email:  ddlc@techresc.net

	This source file defines Filesystem access routines.

	Modification History :

	Version 2.1:
		02.05.2004 : new.
*/

#include "config.h"
#if INCLUDE_FS
#include "stddefs.h"
#include "errno.h"
#include "cli.h"

#include "fs.h"

#if INCLUDE_FAT
#include "fat.h"
#endif

master_boot_record MBR;
uchar sector0[FS_BLOCK_SIZE];
boot_sector BPB;
file_system FS = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};


/* mount_fs function.

	Entry :
		file_system_ops * fsops : filesystem structure to initialise.

	Exit :
		file_system_ops * fsops : filesystem structure to initialise.
		return 0 if successfull, else error code.

	Function :
		It mounts filesystem structures.

*/
int mount_fs(file_system_ops * fsops)
{
	int	ret_val = -ENODEV;

#if INCLUDE_FAT
	// Call init_media function from fat module.
	ret_val = init_media(fsops);

	if(init_sector0(&MBR, (ulong *)sector0) != 0)
	{	// Sector0 is not found.
		printf("FS not found\n");
	}
	else
	{	// MBR or/and boot sector are found, init BPB structure.
		init_boot_sector(&BPB, (ulong *)sector0);
		if(init_file_system(&FS, &BPB)== 0)
		{
			printf("File system found, ");
			if(FS.fat_type == PDOS32)
			{	printf("FAT32.\n");
			}
			else
			{	printf("FAT16.\n");
			}
			ret_val = 0;		// Indicate success.
		}
	}
#endif

	return ret_val;
}

/* fs_write function.

	Entry :
		char * filename : filename in 8.3 format.
		ulong src : 32-bit source address to copy from.
		ulong filelength : file size in bytes.

	Exit :
		<0 : error in file write.
		>0 : bytes written. (should be equal to filelength to validate write.


*/
int fs_write(char * filename, ulong src, ulong filelength)
{
	int retval = -1;						// Assume file write failed.
	file_infos first_fi, fi, next_fi;		// To contain file infos.
	int bytes_written, bytes_remain;

#if INCLUDE_FAT
	// Check file name validity.
	first_fi = find_filename(filename, &FS);
	if(first_fi.cluster_num == 0)
	{	// This file doesn't exist.
#ifdef FAT_DEBUG
		print_file_infos(&first_fi);
#endif

		// Init directory entry.
		init_dir_entry(&first_fi, &FS);	// first_fi will be used at the end of fs_write to update dir entry. 

		// Find first free cluster.
		first_fi = find_free_cluster(&first_fi, &FS);
#ifdef FAT_DEBUG
		print_file_infos(&first_fi);
#endif
		// Init current file info structure with first one.
		fi = first_fi;

		// Update first file info structure to save filelength.
		first_fi.file_size = filelength;

		// We are ready now to write file to data area.
		bytes_remain = filelength;
		while(bytes_remain > 0)
		{	// Exit when bytes_remain <= 0).
			// Write cluster.
#ifdef FAT_DEBUG
			print_file_infos(&fi);
#endif
			bytes_written = write_cluster(&fi, &FS, src);
			if(bytes_written == 0)
			{	// There's nothing else to write.
				break;
			}
			// prepare next write.
			src += bytes_written;
			bytes_remain -= bytes_written;

			if(bytes_remain > 0)
			{
				// Find next free cluster.
				next_fi = find_free_cluster(&fi, &FS);

				// Update fi.current_fat_entry with next_fi.cluster_num.
				update_fat(&fi, &next_fi, &FS);

				// next_fi becomes current fi.
				fi = next_fi;
			}
		}
		if(bytes_remain <= 0)
		{	// File is successfully written.
			// Last FAT entry was already filled in with EOC mark.

			// Update directory entry and write last fat sector if needed.
			first_fi.current_fat_entry = fi.current_fat_entry;	// Be sure first_fi.current_fat_entry is last fat sector
			retval = update_file_system(&first_fi, &FS);
		}

	}
	else
	{	// This file exists.
		printf("This file exists, can't override it.\n");
		retval = -ENOENT;
	}
#endif

	return retval;
}

/* fs_read function.

	Entry :
		ulong dest : destination address to copy to.
		char * filename : filename in 8.3 format.

	Exit :
		<0 : error in file read.
		>0 : bytes read.


*/
int fs_read(ulong dest, char * filename)
{
	int retval = -1;	// Assume file read failed.
	file_infos fi, next_fi;		// To contain file infos.
	int bytes_read, bytes_remain;

#if INCLUDE_FAT

	fi = find_filename(filename, &FS);

	if(fi.cluster_num == 0)
	{	// No such file or directory.
#ifdef FAT_DEBUG
		print_file_infos(&fi);
#endif
		retval = -ENOENT;
	}
	else
	{
#ifdef FAT_DEBUG
		print_file_infos(&fi);
#endif

		bytes_remain = fi.file_size;
		while(bytes_remain >= 0)
		{	// Exit when bytes_remain <= 0.

			bytes_read = read_cluster(&fi, &FS, dest);
			if(bytes_read == 0)
			{	// There's nothing else to read.
				break;
			}
			dest += bytes_read;
			bytes_remain -= bytes_read;
	
			next_fi = find_next_cluster(&fi, &FS);

			if( ((FS.fat_type == PDOS32) && (next_fi.cluster_num == PDOS32_EOC))
			 || ((FS.fat_type == PDOS16) && (next_fi.cluster_num == PDOS16_EOC)) )
			{	// EOC = End of chain.
#ifdef FAT_DEBUG
				printf("End of Chain.\n");
#endif
				retval = fi.file_size;
				break;
			}
			else
			{	// next becomes current.
				fi = next_fi;
			}
		}
	}

#endif

	return retval;
}


char *fsHelp[] = {
	"fs operations",
	"{op} [args]",
	"Ops...",
	"  status",
	"  write {filename} {src} {byte_cnt}",
	"  read {dest} {filename}",
	"Read and Write access return file size in FILESIZE environment variable.",
	"",
	0,
};

/* MmcCmd():
 *	Code that handles the user interface.  See mmcHelp[] above for usage.
 */
int
FsCmd(int argc,char *argv[])
{
	int		snum, ret;
	ulong	dest, src, oints;
	long	bytecnt, rslt;
	char	filename[FILENAME_SIZE];
	char	filesize[FILENAME_SIZE];

	ret = CMD_SUCCESS;

	if (strcmp(argv[1],"status") == 0)
		print_file_system(&FS);
	else if (!strcmp(argv[1],"write")) 
	{
		if (argc == 5)
		{
			strcpy(filename, argv[2]);
			src = strtoul(argv[3],(char **)0,0);
			bytecnt = (long)strtoul(argv[4],(char **)0,0);
			rslt = fs_write(filename,src,bytecnt);
			if ((rslt < 0) || (rslt!=bytecnt))
			{
				printf("Write failed\n");
				ret = CMD_FAILURE;
			}
			else
			{
				sprintf(filesize,"%d",rslt);
				setenv("FILESIZE", filesize);
			}
		}
		else
		{
			ret = CMD_PARAM_ERROR;
		}
	}
	else if (!strcmp(argv[1],"read")) 
	{
		if (argc == 4)
		{
			dest = strtoul(argv[2],(char **)0,0);
			strcpy(filename, argv[3]);
			rslt = fs_read(dest,filename);
			if (rslt < 0)
			{
				printf("Read failed\n");
				ret = CMD_FAILURE;
			}
			else
			{
				sprintf(filesize,"%d",rslt);
				setenv("FILESIZE", filesize);
			}
		}
		else
		{
			ret = CMD_PARAM_ERROR;
		}
	}
	else
	{
		ret = CMD_PARAM_ERROR;
	}

	return(ret);
}

#endif /* INCLUDE_FS */
