//==========================================================================
//
// pxa270_gpio.c
//
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Contributors: 
// Date:         02/05/2002
// Description:  This file contains code to intialize the PXA250 GPIO
//               section as well as set, clear and test functions
//               for manipulating the GPIO bits
//
//
// Modified by Luis Torrico, Cogent Computer Systems, Inc.
//	03.28.2007 : Added all GPIO Settings for CSB726 PXA270 Cogent board.
//--------------------------------------------------------------------------

#include "config.h"
#include "cpuio.h"
#include "stddefs.h"
#include "genlib.h"
#include "pxa270.h"

// Function Prototypes
void GPIO_init(void);
int GPIO_set(int);
int GPIO_clr(int);
int GPIO_tst(int);
int GPIO_in(int);
int GPIO_out(int);

//--------------------------------------------------------------------------
//
// GPIO assignments for the CSB726 are as follows.
// Active indicates what level is considered active for that
// signal, if any.  A/F indicates alternate function mode (0 = none).
//
// GPIO		Direction	Active	A/F	Description
// 0		Input		Low		0	*PWR_INT - Power Interrupt from AVR
// 1		I/O			Low		0	GPIO 0 - used as *IRQ to the processor
// 2		Unused
// 3		I/O			N/A		0	PWR_SCL - Power Manager I2C SCL for ISL6271 Programmable Core Regulator
// 4		I/O			N/A		0	PWR_SDA - Power Manager I2C SDA for ISL6271 Programmable Core Regulator
// 5-8		Unused
// 9		I/O			N/A		0	GPIO 2 - Can be used as Timer Channel Output 0
// 10		I/O			N/A		0	GPIO 3 - Can be used as Timer Channel Output 1
// 11		I/O			Low		0	GPIO 1 - Can be used as *PME
// 12		Input		Low		0	*PIRQ - Touch Controller Interrupt
// 13		Output		N/A		1	SPI0_MOSI - Synchronous Serial Port 0 Transmit
// 14		Output		Low		2	*SPI0_CS0 - Synchronous Serial Port 0 Frame
// 15		Input		Low		0	*EXP_IRQ - Used as Expansion Interrupt from the CPLD
// 16		I/O			N/A		0	GPIO 4 - Can be used as PWM0
// 17		I/O			N/A		0	GPIO 5 - Can be used as PWM1
// 18		Input		Low		1	*EXP_WAIT, External Bus Ready (PXA270 RDY)
// 19		Output		N/A		1	SPI0_CLK - Synchronous Serial Port 0 Clock
// 20		Unused					This GPIO is tied to GPIO 109 so it is unavailable.
// 21		Unused		
// 22		Input		N/A		1	SSP2_EXTCLK - Synchronous Serial Port External Clock for SPI0
// 23-27	Unused					Reserved for Quick Capture Interface
// 28		Input		N/A		1	AC_BCLK - AC97 Bit Clock from Codec
// 29		Input		N/A		1	AC_SDIN - AC97 Serial Data from Codec
// 30		Output		N/A		2	AC_SDOUT - AC97 Serial Data to Codec
// 31		Output		N/A		2	AC_SYNC - AC97 SYNC to Codec
// 32		Output		N/A		2	SD_CLK - MMC Clock
// 33		Output		Low		2	Chip Select 5 - Unused
// 34		Input		N/A		1	U0_RXD - Full Function UART RXD
// 35		Input		N/A		1	U0_CTS - Full Function UART CTS
// 36		Input		N/A		1	U0_DCD - Full Function UART DCD
// 37		Input		N/A		1	U0_DSR - Full Function UART DSR
// 38		Input		N/A		1	U0_RI - Full Function UART RI
// 39		Output		N/A		2	U0_TXD - Full Function UART TXD
// 40		Output		N/A		2	U0_DTR - Full Function UART DTR
// 41		Output		N/A		2	U0_RTS - Full Function UART RTS
// 42		Input		N/A		1	U1_RXD - Bluetooth UART RXD
// 43		Output		N/A		2	U1_TXD - Bluetooth UART TXD
// 44		Input		N/A		1	U1_CTS - Bluetooth UART CTS
// 45		Output		N/A		2	U1_RTS - Bluetooth UART RTS
// 46		Input		N/A		2	D_RXD (Debug UART) - Standard UART RXD
// 47		Output		N/A		1	D_TXD (Debug UART) - Standard UART TXD
// 48		Output		Low		2	*CF_OE - Compact FLASH Output Enable
// 49		Output		Low		2	*CF_WE - Compact FLASH Write Enable
// 50		Output		Low		2	*CF_IOR - Compact FLASH I/O Read Enable	 
// 51		Output		Low		2	*CF_IOW - Compact FLASH I/O Write Enable
// 52		Input		Low		0	*E_INT - LAN9215 Interrupt
// 53		Input		Low		0	*SM_INT - SM501 Interrupt
// 54		Output		Low		2	*CF_CE2 - Compact Flash Chip Enable 2
// 55		Output		Low		2	*CF_REG - Compact Flash Register Select
// 56		Input		Low		1	*CF_WAIT - Compact Flash Wait Request
// 57		Unused
// 58-77	Unused					Reserved for LCD Interface
// 78		Output		Low		2	*EXP_CS - Chip Select 2 - Expansion CS to CPLD
// 79		Output		Low		2	Chip Select 3 - LAN9215
// 80		Output		Low		2	Chip Select 4 - SM501
// 81		Output		N/A		1	SPI1_MOSI - Synchronous Serial Port 1 Transmit
// 82		Input		N/A		1	SPI1_MOSI - Synchronous Serial Port 1 Receive
// 83		Output		Low		1	*SPI1_CS0 - Synchronous Serial Port 1 Frame
// 84		Output		N/A		1	SPI1_CLK - Synchronous Serial Port 1 Clock
// 85		Output		Low		1	*CF_CE1 - Compact Flash Chip Enable 1
// 86		Input		Low		1	SPI0_MOSI - Synchronous Serial Port 0 Receive
// 87		Output		High	0	PWR_1WIRE
// 88		Unused
// 89		Unused
// 90		Unused					Reserved for Quick Capture Interface
// 91		Unused					Reserved for Quick Capture Interface
// 92		I/O			N/A		1	SD_D0 - MMC Data Bit 0
// 93		Unused					Reserved for Quick Capture Interface
// 94		I/O 		N/A		0	GPIO 6
// 95		I/O 		N/A		0	GPIO 7
// 96		I/O 		N/A		0	GPIO 8
// 97		I/O 		N/A		0	GPIO 9
// 98		Input		High	0	CF_RDY - Compact FLASH Card Ready/Busy and Interrupt
// 99		Input		Low		0	*CF_CD - Compact FLASH Card Detect
// 100		Input		Low		0	*SD_CD - MMC Card Detect
// 101		Input		Low		0	*SD_WP - MMC Write Protect
// 102		Unused
// 103		Output		High	0	CF_RST - Compact Flash Card Reset
// 104		Output		Low		0	PWR_OFF - Power Off request to AVR
// 105		Unused
// 106		Unused
// 107		Unused
// 108		Unused					Reserved for Quick Capture Interface
// 109		I/O 		N/A		1	SD_D1 - MMC Data Bit 1 (MMC Interrupt)
// 110		I/O 		N/A		1	SD_D2 - MMC Data Bit 2
// 111		I/O 		N/A		1	SD_D3 - MMC Data Bit 3
// 112		Output		N/A		1	SD_CMD - MMC Command
// 113		Output		Low		2	*AC_RST - AC97 Reset to Codec
// 114		Unused					Reserved for Quick Capture Interface
// 115		Unused					Reserved for Quick Capture Interface
// 116		Unused					Reserved for Quick Capture Interface
// 117		I/O 		N/A		1	I2C_SCL - I2C Clock
// 118		I/O 		N/A		1	I2C_SDA - I2C Data

//--------------------------------------------------------
// GPIO_init()
//
// This function sets the startup state for the PXA270
// as used on the CSB726
//
void
GPIO_init()
{

	// First set the output values to a safe/disabled state before we change any GPIO's to outputs
	// start by setting all of them high which is the safest for most signals
	GPIO_REG(GPIO_GPSR0) = 0xffffffff;
	GPIO_REG(GPIO_GPSR1) = 0xffffffff;
	GPIO_REG(GPIO_GPSR2) = 0xffffffff;
	GPIO_REG(GPIO_GPSR3) = 0xffffffff;

	// Now clear any high true outputs that need to start low
	GPIO_REG(GPIO_GPCR0) = 0;			// 
	
	GPIO_REG(GPIO_GPCR1) = 0;			// 

	GPIO_REG(GPIO_GPCR2) = 0;			// 

	GPIO_REG(GPIO_GPCR3) = GPIO_103;		// CF_RST

	// Next, set the correct direction for our functions.  A one means output.
	GPIO_REG(GPIO_GPDR0) = GPIO_9 		// CHOUT0 (GPIO 2)
  						 | GPIO_10 		// CHOUT1 (GPIO 3)
  						 | GPIO_11 		// PME (GPIO 1)
  						 | GPIO_13		// SSP2_TXD
  						 | GPIO_14		// SSP2_FRM
  						 | GPIO_16		// PWM0 (GPIO 4)
  						 | GPIO_17		// PWM1 (GPIO 5)
  						 | GPIO_19		// SSP2_CLK
  						 | GPIO_30		// AC97 Serial Data to Codec
  						 | GPIO_31;		// AC97 SYNC to Codec

	GPIO_REG(GPIO_GPDR1) = GPIO_32		// MMC_CLK
  						 | GPIO_33		// Chip Select 5 - Unused
  						 | GPIO_39		// Full Function UART TXD
  						 | GPIO_40		// Full Function UART DTR
  						 | GPIO_41		// Full Function UART RTS
  						 | GPIO_43		// Bluetooth UART TXD
  						 | GPIO_45		// Bluetooth UART RTS
  						 | GPIO_47		// Standard/IrDA UART TXD
  						 | GPIO_48		// Compact FLASH OE
  						 | GPIO_49		// Compact FLASH WE
  						 | GPIO_50		// Compact FLASH IOR
  						 | GPIO_51		// Compact FLASH IOW
  						 | GPIO_54		// Compact FLASH CE2
  						 | GPIO_55		// Compact FLASH REG
  						 | GPIO_58		// LCD Data 0
  						 | GPIO_59		// LCD Data 1
  						 | GPIO_60		// LCD Data 2
  						 | GPIO_61		// LCD Data 3
  						 | GPIO_62		// LCD Data 4
  						 | GPIO_63;		// LCD Data 5

	GPIO_REG(GPIO_GPDR2) = GPIO_64		// LCD Data 6
  						 | GPIO_65		// LCD Data 7
  						 | GPIO_66		// LCD Data 8
  						 | GPIO_67		// LCD Data 9
  						 | GPIO_68		// LCD Data 10
  						 | GPIO_69		// LCD Data 11
  						 | GPIO_70		// LCD Data 12
  						 | GPIO_71		// LCD Data 13
  						 | GPIO_72		// LCD Data 14
  						 | GPIO_73		// LCD Data 15
  						 | GPIO_74		// LCD FCLK (VSYNC)
  						 | GPIO_75		// LCD LCLK (HSYNC)
  						 | GPIO_76		// LCD PCLK
  						 | GPIO_77		// LCD ACBIAS, Active LCD Enable
  						 | GPIO_78		// Chip Select 2 - Expansion
  						 | GPIO_79		// Chip Select 3 - LAN9115
  						 | GPIO_80		// Chip Select 4 - SM501   
  						 | GPIO_81		// SSP3_TXD
  						 | GPIO_83		// SSP3_FRM
  						 | GPIO_84		// SSP3_CLK
  						 | GPIO_85		// Compact FLASH CE1
  						 | GPIO_87		// 1-Wire
  						 | GPIO_89		// USB Host 1 Enable
  						 | GPIO_94		// GPIO 6
  						 | GPIO_95;		// GPIO 7


	GPIO_REG(GPIO_GPDR3) = GPIO_96		// GPIO 8
  						 | GPIO_97		// GPIO 9
  						 | GPIO_98		// CF_RDY
  						 | GPIO_103		// CF Reset
  						 | GPIO_113		// AC97 Reset
  						 | GPIO_117;	// SCL

	// Finally, set the alternate function registers to the correct state
	GPIO_REG(GPIO_GAFR0L) = GPIO_13_AF_SSP2_TXD			// AF1
						  | GPIO_14_AF_SSP2_FRM;		// AF2

	GPIO_REG(GPIO_GAFR0U) = GPIO_18_AF_RDY				// AF1	External Bus Ready (used by CS8900a)
				 		  | GPIO_19_AF_SSP2_CLK		 	// AF1	
				 		  | GPIO_22_SSP_EXTCLK		 	// AF1	
				 		  | GPIO_28_AF_AC97_BCLK_IN 	// AF1	AC97 Bit Clock from Codec
				 		  | GPIO_29_AF_AC97_SDIN0	 	// AF1	AC97 Serial Data from Codec
				 		  | GPIO_30_AF_AC97_SDOUT		// AF2	AC97 Serial Data to Codec
				 		  | GPIO_31_AF_AC97_SYNC;		// AF2	AC97 SYNC to Codec

	GPIO_REG(GPIO_GAFR1L) = GPIO_32_AF_MMC_CLK			// AF2
						  | GPIO_33_AF_CS5				// AF2	Chip Select 5 - TD242hc USB HOST
				  		  | GPIO_34_AF_FF_RXD			// AF1	Full Function UART RXD
				  		  | GPIO_35_AF_FF_CTS			// AF1	Full Function UART RXD
				  		  | GPIO_36_AF_FF_DCD			// AF1	Full Function UART RXD
				  		  | GPIO_37_AF_FF_DSR			// AF1	Full Function UART RXD
				  		  | GPIO_38_AF_FF_RI			// AF1	Full Function UART RXD
				  		  | GPIO_39_AF_FF_TXD			// AF2	Full Function UART TXD
				  		  | GPIO_40_AF_FF_DTR			// AF2	Full Function UART TXD
				  		  | GPIO_41_AF_FF_RTS			// AF2	Full Function UART TXD
				  		  | GPIO_42_AF_BT_RXD			// AF1	Blue Tooth UART TXD
				  		  | GPIO_43_AF_BT_TXD			// AF2	Blue Tooth UART TXD
				  		  | GPIO_44_AF_BT_CTS			// AF1	Blue Tooth UART TXD
				  		  | GPIO_45_AF_BT_RTS			// AF2	Blue Tooth UART TXD
				  		  | GPIO_46_AF_STD_RXD			// AF2	Standard UART RXD
				  		  | GPIO_47_AF_STD_TXD;			// AF1	Standard UART TXD
          
	GPIO_REG(GPIO_GAFR1U) = GPIO_48_AF_POE				// AF2	Compact FLASH OE
				  		  | GPIO_49_AF_PWE				// AF2	Compact FLASH WE
				  		  | GPIO_50_AF_PIOR				// AF2	Compact FLASH IOR
				  		  | GPIO_51_AF_PIOW				// AF2	Compact FLASH IOW
				  		  | GPIO_54_AF_PCE2				// AF2	Compact FLASH CE1
				  		  | GPIO_55_AF_PREG				// AF2	Compact FLASH REG
				  		  | GPIO_56_AF_PWAIT			// AF1	Compact FLASH WAIT
				  		  | GPIO_58_AF_LDD0				// AF2	LCD Data 0
				  		  | GPIO_59_AF_LDD1				// AF2	LCD Data 1
				  		  | GPIO_60_AF_LDD2				// AF2	LCD Data 2
				  		  | GPIO_61_AF_LDD3				// AF2	LCD Data 3
				  		  | GPIO_62_AF_LDD4				// AF2	LCD Data 4
				  		  | GPIO_63_AF_LDD5;			// AF2	LCD Data 5

	GPIO_REG(GPIO_GAFR2L) = GPIO_64_AF_LDD6				// AF2	LCD Data 6
						  | GPIO_65_AF_LDD7				// AF2	LCD Data 7
						  | GPIO_66_AF_LDD8				// AF2	LCD Data 8
						  | GPIO_67_AF_LDD9				// AF2	LCD Data 9
						  | GPIO_68_AF_LDD10			// AF2	LCD Data 10
						  | GPIO_69_AF_LDD11			// AF2	LCD Data 11
						  | GPIO_70_AF_LDD12			// AF2	LCD Data 12
						  | GPIO_71_AF_LDD13			// AF2	LCD Data 13
						  | GPIO_72_AF_LDD14			// AF2	LCD Data 14
						  | GPIO_73_AF_LDD15			// AF2	LCD Data 15
						  | GPIO_74_AF_LCD_FCLK			// AF2	LCD FCLK (VSYNC)
						  | GPIO_75_AF_LCD_LCLK			// AF2	LCD LCLK (HSYNC)
						  | GPIO_76_AF_LCD_PCLK			// AF2	LCD PCLK
						  | GPIO_77_AF_LCD_BIAS			// AF2	LCD ACBIAS (LCD ON)
						  | GPIO_78_AF_CS2				// AF2	Chip Select 2 - expansion
						  | GPIO_79_AF_CS3;				// AF2	Chip Select 3 - LAN9115

	GPIO_REG(GPIO_GAFR2U) = GPIO_80_AF_CS4				// AF2	Chip Select 4 - SM501
						  | GPIO_81_AF_SSP3_TXD			// AF2	
						  | GPIO_82_AF_SSP3_RXD			// AF1	
						  | GPIO_83_AF_SSP3_FRM			// AF1	
						  | GPIO_84_AF_SSP3_CLK			// AF1	
						  | GPIO_85_AF_PCE1				// AF1
						  | GPIO_86_AF_SSP2_RXD			// AF1	
						  | GPIO_92_AF_MMC_D0_I;		// AF2	

  	GPIO_REG(GPIO_GAFR3L) = GPIO_109_AF_MMC_D1_I		// AF1
						  | GPIO_110_AF_MMC_D2_I		// AF1	
						  | GPIO_111_AF_MMC_D3_I;		// AF1	

  	GPIO_REG(GPIO_GAFR3U) = GPIO_112_AF_MMC_CMD_I		// AF1
						  | GPIO_113_AF_AC97_RST		// AF2	
						  | GPIO_117_AF_I2C_SCL_O		// AF1	
						  | GPIO_118_AF_I2C_SDA_I;		// AF1	

	// finally, clear the RDH bit (by writing 1 to it) in the PSSR to enable GPIO inputs
	PWR_REG(PWR_PSSR) = 0x20;
}

//--------------------------------------------------------
// GPIO_set()
//
// This function sets the desired bit passed in.  It first 
// tests to see if the bit is an actual output.  If not,
// it returns a -1, otherwise it returns 0.  NOTE: We do not
// test to see if setting the bit would screw up any alternate
// functions.  Use this function with caution!
//

int GPIO_set(int gpio_bit)
{
	// quick sanity test
	if (gpio_bit > 118) return -1;

	// First find out which group the bit belongs to
	if (gpio_bit < 32)
	{
		// Get the direction and see if it's an output
		if (!(GPIO_REG(GPIO_GPDR0) & (1 << gpio_bit)))
			return -1;

		// if yes, then set it
		GPIO_REG(GPIO_GPSR0) = (1 << gpio_bit);
	}
	else if (gpio_bit < 64)
	{
		// Get the direction and see if it's an output
		if (!(GPIO_REG(GPIO_GPDR1) & (1 << (gpio_bit - 32))))
			return -1;

		// if yes, then set it
		GPIO_REG(GPIO_GPSR1) = (1 << (gpio_bit - 32));
	}
	else if (gpio_bit < 96)
	{
		// Get the direction and see if it's an output
		if (!(GPIO_REG(GPIO_GPDR2) & (1 << (gpio_bit - 64))))
			return -1;

		// if yes, then set it
		GPIO_REG(GPIO_GPSR2) = (1 << (gpio_bit - 64));
	}
	else
	{
		// Get the direction and see if it's an output
		if (!(GPIO_REG(GPIO_GPDR3) & (1 << (gpio_bit - 96))))
			return -1;

		// if yes, then set it
		GPIO_REG(GPIO_GPSR3) = (1 << (gpio_bit - 96));
	}
	return 0;
}

//--------------------------------------------------------
// GPIO_clr()
//
// This function clears the desired bit passed in.  It first 
// tests to see if the bit is an actual output.  If not,
// it returns a -1, otherwise it returns 0.  NOTE: We do not
// test to see if clearing the bit would screw up any alternate
// functions.  Use this function with caution!
//

int GPIO_clr(int gpio_bit)
{
	// quick sanity test
	if (gpio_bit > 118) return -1;

	// First find out which group the bit belongs to
	if (gpio_bit < 32)
	{
		// Get the direction and see if it's an output
		if (!(GPIO_REG(GPIO_GPDR0) & (1 << gpio_bit)))
			return -1;

		// if yes, then set it
		GPIO_REG(GPIO_GPCR0) = (1 << gpio_bit);
	}
	else if (gpio_bit < 64)
	{
		// Get the direction and see if it's an output
		if (!(GPIO_REG(GPIO_GPDR1) & (1 << (gpio_bit - 32))))
			return -1;

		// if yes, then clear it
		GPIO_REG(GPIO_GPCR1) = (1 << (gpio_bit - 32));
	}
	else if (gpio_bit < 96)
	{
		// Get the direction and see if it's an output
		if (!(GPIO_REG(GPIO_GPDR2) & (1 << (gpio_bit - 64))))
			return -1;

		// if yes, then clear it
		GPIO_REG(GPIO_GPCR2) = (1 << (gpio_bit - 64));
	}
	else
	{
		// Get the direction and see if it's an output
		if (!(GPIO_REG(GPIO_GPDR3) & (1 << (gpio_bit - 96))))
			return -1;

		// if yes, then clear it
		GPIO_REG(GPIO_GPCR3) = (1 << (gpio_bit - 96));
	}
	return 0;
}
//--------------------------------------------------------
// GPIO_tst()
//
// This function returns the state of desired bit passed in.
// It does not test to see if it's an input or output and thus
// can be used to verify if an output set/clr has taken place
// as well as for testing an input state.
//

int GPIO_tst(int gpio_bit)
{
	// First find out which group the bit belongs to
	if (gpio_bit < 32)
	{
		// Get the state
		if (GPIO_REG(GPIO_GPLR0) & (1 << gpio_bit)) return 1;
	}
	else if (gpio_bit < 64)
	{
		// Get the state
		if (GPIO_REG(GPIO_GPLR1) & (1 << (gpio_bit - 32))) return 1;
	}
	else if (gpio_bit < 96)
	{
		// Get the state
		if (GPIO_REG(GPIO_GPLR2) & (1 << (gpio_bit - 64))) return 1;
	}
	else
	{
		// Get the state
		if (GPIO_REG(GPIO_GPLR3) & (1 << (gpio_bit - 96))) return 1;
	}
	return 0; // bit was not set
}

//--------------------------------------------------------
// GPIO_in()
//
// This function changes the direction of the desired bit 
// to input.  NOTE: We do not test to see if changing the
// direction of the bit would screw up anything.  Use this
// function with caution!
int GPIO_in(int gpio_bit)
{
	// quick sanity test
	if (gpio_bit > 118) return -1;

	// now find out which group the bit belongs to
	if (gpio_bit < 32)
	{
		GPIO_REG(GPIO_GPDR0) &= ~(1 << gpio_bit);
	}
	else if (gpio_bit < 64)
	{
		GPIO_REG(GPIO_GPDR1) &= ~(1 << (gpio_bit - 32));
	}
	else if (gpio_bit < 96)
	{
		GPIO_REG(GPIO_GPDR2) &= ~(1 << (gpio_bit - 64));
	}
	else
	{
		GPIO_REG(GPIO_GPDR3) &= ~(1 << (gpio_bit - 96));
	}
	return 0;
}

//--------------------------------------------------------
// GPIO_out()
//
// This function changes the direction of the desired bit 
// to output.  NOTE: We do not test to see if changing the
// direction of the bit would screw up anything.  Use this
// function with caution!
int GPIO_out(int gpio_bit)
{
	// quick sanity test
	if (gpio_bit > 118) return -1;

	// now find out which group the bit belongs to
	if (gpio_bit < 32)
	{
		GPIO_REG(GPIO_GPDR0) |= (1 << gpio_bit);
	}
	else if (gpio_bit < 64)
	{
		GPIO_REG(GPIO_GPDR1) |= (1 << (gpio_bit - 32));
	}
	else if (gpio_bit < 96)
	{
		GPIO_REG(GPIO_GPDR2) |= (1 << (gpio_bit - 64));
	}
	else
	{
		GPIO_REG(GPIO_GPDR3) |= (1 << (gpio_bit - 96));
	}
	return 0;
}

